/*
 * gpcmescc.c: Driver for the GPCM ESCC DMA Card.
 * Copyright (c) 1998,1999,2000 GPCM
 *
 * Heavily based on pi2.c by David Perry.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2, as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software Foundation,
 *  Inc., 675 Mass Ave, Cambridge MA 02139, USA.
 *
 *  The file skeleton.c by Donald Becker was used as a starting point
 *  for this driver.
 *
 *  				   *   *   *
 *
 *    GPCM ESCC card I/O ports allocations:
 *    --------------------------------------------------------------------------
 *    Base+0  (e.g. 0x380) ESCC channel B control register
 *    Base+1  (e.g. 0x381) ESCC channel B data register
 *    Base+2  (e.g. 0x382) ESCC channel A control register
 *    Base+3  (e.g. 0x383) ESCC channel A data register
 *    Base+4  (e.g. 0x384) DMA & IRQ enable register:
 *    bit 0: IRQ enable (1 enabled 0 disabled)
 *    bit 1: DMA enable (1 enabled 0 disabled)
 *    Base+8  (e.g. 0x388) 8254 Timer 0
 *    Base+9  (e.g. 0x389) 8254 Timer 1
 *    Base+10 (e.g. 0x38a) 8254 Timer 2
 *    Base+11 (e.g. 0x38b) 8254 control register
 *
 *    Timer 0 is used as clock source for timers 1 and 2
 *    (OUT0 is connected to CLK1 and CLK2).
 *
 *    8254 Out1 & Out2 go respectively to ESCC CTSA & CTSB pin.
 *    When a timer expires an External Interrupt is generated.
 *
 *    Channel A is DMA driven, channel B interrupt driven.
 *
 *    See card schematic diagram for details.
 *    ---------------------------------------------------------------------------
 */

#ifndef MODULE
#error ******** This driver must be compiled as a module ********
#else

static const char *version = "GPCM ESCC DMA Card Linux driver v0.8 (c) 1998,1999,2000 GPCM\n";

/* 
 * The default configuration 
 */
#define DEF_A_SPEED 38400	/* 38400 baud, 0 = external clock */
#define DEF_A_TXDELAY 500	/* 50 mS transmit delay */
#define DEF_A_PERSIST 128	/* 50% persistence */
#define DEF_A_SLOTIME 500	/* 50 mS slot time */
#define DEF_A_SQUELDELAY 50	/* 5 mS squelch delay - allows fcs and flag */
#define DEF_A_CLOCKMODE 0	/* clock mode - 0 is normal */

#define DEF_B_SPEED 38400	/* 38400 baud */
#define DEF_B_TXDELAY 500	/* 50 mS */
#define DEF_B_PERSIST 128	/* 50% */
#define DEF_B_SLOTIME 500	/* 50 mS */
#define DEF_B_SQUELDELAY 50	/* 5 mS */
#define DEF_B_CLOCKMODE 0	/* Normal clock mode */

#include <linux/module.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/if_arp.h>
#include <linux/in.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/mm.h>
#include <linux/netdevice.h>
#include <linux/sockios.h>
#include <linux/tqueue.h>
#include <linux/version.h>
#include <asm/atomic.h>
#include <asm/bitops.h>
#include <asm/dma.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/segment.h>
#include <asm/uaccess.h>
#include <net/ax25.h>
#include "gpcmescc.h"
#include "z85230.h"
#include "z8254.h"

struct mbuf {
	struct mbuf *next;
	int cnt;
	char data[0];
};

/* The number of low I/O ports used by the card. */
#define GPCM_TOTAL_SIZE	16

/* Index to functions, as function prototypes. */
static void escc_init(struct device *dev);
int gpcm_probe(struct device *dev);
int gpcm_mod_probe(struct device *dev);
static int gpcm_init(struct device *dev, int card_type);
static void reset_ESCC(struct device *dev);
static int gpcm_open(struct device *dev);
static int gpcm_send_packet(struct sk_buff *skb, struct device *dev);
static void gpcm_interrupt(int reg_ptr, void *dev_id, struct pt_regs *regs);
static int gpcm_close(struct device *dev);
static int gpcm_ioctl(struct device *dev, struct ifreq *ifr, int cmd);
static struct enet_statistics *gpcm_get_stats(struct device *dev);
static void rts(struct gpcm_local *lp, int x);
static inline void b_rxint(struct device *dev, struct gpcm_local *lp);
static inline void b_rxflush(struct gpcm_local *);
static inline void b_txint(struct gpcm_local *lp);
static inline void b_exint(struct gpcm_local *lp);
static inline void a_rxint(struct device *dev, struct gpcm_local *lp);
static inline void a_txint(struct gpcm_local *lp);
static inline void a_exint(struct gpcm_local *lp);
static char *get_dma_buffer(unsigned long *mem_ptr);
static int valid_dma_page(unsigned long addr, unsigned long dev_buffsize);
static void tdelay(struct gpcm_local *lp, int time);
static int gpcm_set_mac_address(struct device *dev, void *addr);

static char ax25_bcast[7] = { 'Q' << 1, 'S' << 1, 'T' << 1, ' ' << 1, ' ' << 1, ' ' << 1,
	'0' << 1
};
static char ax25_test[7] = { 'L' << 1, 'I' << 1, 'N' << 1, 'U' << 1, 'X' << 1, ' ' << 1,
	'1' << 1
};

static struct device dev_gpcm0a = { "gpcm0a\0\0\0\0", 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, gpcm_probe };
static struct device dev_gpcm0b = { "gpcm0b\0\0\0\0", 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, gpcm_probe };

static int ext2_secrm_seed = 152;	/* Random generator base */

static inline unsigned char random(void)
{
	return (unsigned char) (ext2_secrm_seed = ext2_secrm_seed * 69069l + 1);
}

static inline void wrtscc(int cbase, int ctl, int sccreg, int val)
{
	outb_p(DMAON, cbase + DMAIRQEN);	/* Disable IRQ while we touch the scc */
	outb_p(sccreg, ctl);	/* Select register */
#ifdef IODELAY
	udelay(IODELAY);
#endif
	outb_p(val, ctl);	/* Output value */
#ifdef IODELAY
	udelay(IODELAY);
#endif
	outb_p(DMAIRQON, cbase + DMAIRQEN);	/* Enable DMA & IRQ */
}

static inline int rdscc(int cbase, int ctl, int sccreg)
{
	int retval;

	outb_p(DMAON, cbase + DMAIRQEN);	/* Disable DMA & IRQ while we touch the scc */
	outb_p(sccreg, ctl);	/* Select register */
#ifdef IODELAY
	udelay(IODELAY);
#endif
	retval = inb_p(ctl);
#ifdef IODELAY
	udelay(IODELAY);
#endif
	outb_p(DMAIRQON, cbase + DMAIRQEN);	/* Enable DMA & IRQ */
	return retval;
}

static void switchbuffers(struct gpcm_local *lp)
{
	if (lp->rcvbuf == lp->rxdmabuf1)
		lp->rcvbuf = lp->rxdmabuf2;
	else
		lp->rcvbuf = lp->rxdmabuf1;
}

/* 
 * Probe for a GPCM ESCC card. Called by gpcm_probe().
 * This routine also initializes the timer chip and issues
 * a reset to the ESCC.
 */
static int hw_probe(int ioaddr)
{
	int time = 10000;	/* Number of 100 us units for test (1s) */
	unsigned long start_time, end_time, njiffies;
	int base, tmr0, tmr1, tmr2, tmrcmd, status1, status2, ctla, ctlb;
	int have_8254 = 0, have_scc = 0, have_escc = 0;
	int a = 1;
	int b = 1;
	int retries = 0;	/* if something fails, reread timer */

	base = ioaddr & 0x3f0;
	tmr0 = TMR0 + base;
	tmr1 = TMR1 + base;
	tmr2 = TMR2 + base;
	tmrcmd = TMRCMD + base;
	while (retries < 3) {
		/* 
		 * Set up counter chip timer 0 for 100 uS period square wave
		 * assuming a 4.9152 Mhz clock.
		 */
		outb_p(SC0 | LSB_MSB | MODE3, tmrcmd);
		udelay(IODELAY);
		outb_p(DIV0 & 0xFF, tmr0);
		udelay(IODELAY);
		outb_p(DIV0 >> 8, tmr0);
		/* 
		 * Setup timer control word for timer 1 and 2
		 */
		outb_p(SC1 | LSB_MSB | MODE0, tmrcmd);
		outb_p(time & 0xFF, tmr1);
		outb_p((time >> 8) & 0xFF, tmr1);
		outb_p(SC2 | LSB_MSB | MODE0, tmrcmd);
		outb_p(time & 0xFF, tmr2);
		outb_p((time >> 8) & 0xFF, tmr2);
		/* 
		 * wait until Timer 2 counter reg is loaded
		 */
		start_time = jiffies;
		do {
			/* Latch count for reading */
			outb_p(SC2, tmrcmd);
			a = inb_p(tmr2);
			b = inb_p(tmr2);
			end_time = jiffies;
			/* Don't wait forever - there may be no card here */
			if ((end_time - start_time) > 200)
				return 0;	/* No card found */
		}
		while (b != (time >> 8));
#if GPCM_DEBUG > 1
		printk("0x%x%x read from timer, start test\n", b, a);
#endif
		start_time = jiffies;
		while (b != 0) {
			/* Latch count for reading */
			outb_p(SC2, tmrcmd);
			a = inb_p(tmr2);
			b = inb_p(tmr2);
#if GPCM_DEBUG > 1
			if (b == 0xff)
				printk("GPCM: Warning 0xFF read from timer\n");
#endif
			end_time = jiffies;
			/* Don't wait forever - there may be no card here */
			if ((end_time - start_time) > 200)
				return 0;	/* No card found */
		}
#if GPCM_DEBUG > 1
		printk("0x%x%x read from timer, end test\n", b, a);
#endif
		end_time = jiffies;
		njiffies = end_time - start_time;
		/* For a 4.9152 MHz clock, njiffies = 97 */
#if GPCM_DEBUG > 0
		printk("GPCM DEBUG: hw_probe(): counted %lu jiffies\n", njiffies);
#endif
		if (njiffies > 80)
			break;
		else
			printk("GPCM: retry to read clock (%d)\n", ++retries);
	};

	if (njiffies > 80) {
		/* 
		 * There is a timer chip, let's see if it is a 8254:
		 * try to read status using the "Read Back" feature.
		 *
		 * Mode read must be equal to the mode selected above.
		 * That works ONLY with a 8254 chip.
		 *
		 * Status is read two times to ensure 8253s are always
		 * correctly detected (on 8253s, counter 0 is actually
		 * read, not status, so reading it only once status may
		 * be equal to mode by chance, resulting in a false 8254
		 * detection.
		 */
		outb_p(RDBACK | RDBSTA | RDBSC0, tmrcmd);
		status1 = inb_p(tmr0);
		outb_p(RDBACK | RDBSTA | RDBSC0, tmrcmd);
		status2 = inb_p(tmr0);
#if GPCM_DEBUG > 1
		printk("GPCM DEBUG: status1=0x%2.2x, status2=0x%2.2x\n", status1, status2);
#endif
		if (((status1 & MODE) == (LSB_MSB | MODE3))
		    && ((status2 & MODE) == (LSB_MSB | MODE3))) {
			/* Mode reads back OK, this is a 8254 or compatible */
			have_8254 = 1;
		} else {
			/* This is a 8253 or compatible */
			have_8254 = 0;
		}
	} else {
		printk("GPCM WARNING: timer chip clock != 4.9152 MHz !\n");
		printk("              measured clock is %ld KHz\n", ((DIV0 * 1000) / njiffies));
		return 4;	/* Non-standard clock found */
	}
	/*
	 * Now check if the card has a 85230 ESCC chip.
	 */
	ctla = base + CHANA + CTL;	/* ESCC channel A control register */
	ctlb = base + CHANB + CTL;	/* ESCC channel B control register */
	/*
	 * Reset 85230
	 */
	outb_p(R9, ctla);
	outb_p(FHWRES, ctla);
	/*
	 * Allow hardware reset completion
	 */
	udelay(20);
	/*
	 * Disable interrupts
	 */
	outb_p(R9, ctla);
	outb_p(0, ctla);
	/*
	 * check for any SCC
	 */
	outb_p(R15, ctlb);
	outb_p(0, ctlb);
	outb_p(R15, ctlb);
	have_scc = ((inb_p(ctlb) & ESCC) == 0);
	/*
	 * Check for ESCC
	 */
	if (have_scc) {
		outb_p(R15, ctlb);
		outb_p(ESCC, ctlb);
		outb_p(R15, ctlb);
		have_escc = (inb_p(ctlb) & ESCC);
		outb_p(R15, ctlb);
		outb_p(0, ctlb);
	}
	return have_8254 + 2 * have_escc;
}

/*
 * Check for ESCC card existence.
 * Rely on parameters passed by insmod since autoprobing
 * in modules is dangerous.
 */
int gpcm_probe(struct device *dev)
{
	static int first_time = 1;
	static int card_type = 0;
	int ioaddr = dev->base_addr & 0x3f0;

#if GPCM_DEBUG > 0
	printk
	    ("gpcm_probe() called for %s with dev=%p, dev->next=%p, ioaddr=%x, irq=%d, dma=%d\n",
	     dev->name, dev, dev->next, ioaddr, (int) dev->irq, (int) dev->dma);
#endif
	if (first_time) {
		first_time = 0;
		printk(version);
		printk("GPCM: Checking for ESCC DMA card at 0x%x.\n", ioaddr);
		if (check_region(ioaddr, GPCM_TOTAL_SIZE) == 0) {
			card_type = hw_probe(ioaddr);
		}
		switch (card_type) {
		case 2:
			printk("GPCM: Found a GPCM ESCC card with 8253 at address %#3x\n",
			       (int) dev->base_addr & 0x3f0);
			break;
		case 3:
			printk("GPCM: Found a GPCM ESCC card with 8254 at address %#3x\n",
			       (int) dev->base_addr & 0x3f0);
			break;
		case 4:
			printk
			    ("GPCM: Found a card with non-standard timer chip clock at address %#3x\n",
			     (int) dev->base_addr & 0x3f0);
			break;
		default:
			printk("GPCM: ERROR: No GPCM ESCC card and/or no ESCC found\n");
			return -EIO;
		}
	}
	/* 
	 * Allocate space for 4 buffers even though we only need 3,
	 * because one of them may cross a DMA page boundary and
	 * be rejected by get_dma_buffer().
	 */
	dev->priv =
	    kmalloc(sizeof(struct gpcm_local) + (DMA_BUFF_SIZE + sizeof(struct mbuf)) * 4 + 7,
		    GFP_KERNEL | GFP_DMA);
#if GPCM_DEBUG > 1
	if (dev->priv == NULL)
		printk("GPCM DEBUG: kmalloc() returned NULL !\n");
#endif
	/* Now initialize them */
	return gpcm_init(dev, card_type);
}

/*
 * Init various data structures and autodetect IRQ and DMA.
 * Called from gpcm_probe() if probe was successful.
 */
static int gpcm_init(struct device *dev, int card_type)
{
	short ioaddr;
	struct gpcm_local *lp;
	unsigned long mem_ptr;

	ioaddr = dev->base_addr;

	/* Initialize the device structure. */
	/* Must be done before reset_ESCC */
	/* Make certain the data structures used by the GPCM are aligned. */
	dev->priv = (void *) (((int) dev->priv + 7) & ~7);
	lp = (struct gpcm_local *) dev->priv;

	memset(dev->priv, 0, sizeof(struct gpcm_local));

	/* Allocate some buffers which do not cross DMA page boundaries */
	mem_ptr = (unsigned long) dev->priv + sizeof(struct gpcm_local);
	lp->txdmabuf = get_dma_buffer(&mem_ptr);
	lp->rxdmabuf1 = (struct mbuf *) get_dma_buffer(&mem_ptr);
	lp->rxdmabuf2 = (struct mbuf *) get_dma_buffer(&mem_ptr);

	/* Initialize rx buffer */
	lp->rcvbuf = lp->rxdmabuf1;
	lp->rcp = lp->rcvbuf->data;
	lp->rcp[0] = lp->rcvbuf->cnt = 0;

	/* Initialize the transmit queue head structure */
	skb_queue_head_init(&lp->sndq);

	/* These need to be initialized before escc_init is called. */
	lp->xtal = (unsigned long) CLOCK;
	lp->base = dev->base_addr;
	lp->cardbase = dev->base_addr & 0x3f0;
	if (dev->base_addr & CHANA) {
		lp->speed = DEF_A_SPEED;
		/* default channel access Params */
		lp->txdelay = DEF_A_TXDELAY;
		lp->persist = DEF_A_PERSIST;
		lp->slotime = DEF_A_SLOTIME;
		lp->squeldelay = DEF_A_SQUELDELAY;
		lp->clockmode = DEF_A_CLOCKMODE;
	} else {
		lp->speed = DEF_B_SPEED;
		/* default channel access Params */
		lp->txdelay = DEF_B_TXDELAY;
		lp->persist = DEF_B_PERSIST;
		lp->slotime = DEF_B_SLOTIME;
		lp->squeldelay = DEF_B_SQUELDELAY;
		lp->clockmode = DEF_B_CLOCKMODE;
	}
	lp->bufsiz = DMA_BUFF_SIZE;
	lp->tstate = IDLE;

	/* Reset ESCC card */
	reset_ESCC(dev);

	if (dev->base_addr & CHANA) {	/* Do these things only for the A port */
		printk("GPCM: using IRQ %d\n", dev->irq);
		/* Grab the region */
		request_region(ioaddr & 0x3f0, GPCM_TOTAL_SIZE, "gpcm");
		printk("GPCM: using DMA %d\n", dev->dma);
		lp->dmachan = dev->dma;
	}

	/* if Channel A */
	/* Fill in the fields of the device structure */
	dev->open = gpcm_open;
	dev->stop = gpcm_close;
	dev->do_ioctl = gpcm_ioctl;
	dev->hard_start_xmit = gpcm_send_packet;
	dev->get_stats = gpcm_get_stats;

	dev->hard_header = ax25_encapsulate;
	dev->rebuild_header = ax25_rebuild_header;
	dev->set_mac_address = gpcm_set_mac_address;

	dev->type = ARPHRD_AX25;	/* AF_AX25 device */
	dev->hard_header_len = 73;	/* We do digipeaters now */
	dev->mtu = 1500;	/* eth_mtu is the default */
	dev->addr_len = 7;	/* sizeof an ax.25 address */
	dev->tx_queue_len = 64;
	dev_init_buffers(dev);

	memcpy(dev->broadcast, ax25_bcast, 7);
	memcpy(dev->dev_addr, ax25_test, 7);

	/* New-style flags. */
	dev->flags = 0;

	return 0;
}

/*
 * Issue a hardware reset to ESCC, reprogram timer chip and disable interrupts
 */
static void reset_ESCC(struct device *dev)
{
	int cardbase;
	unsigned long flags;

#if GPCM_DEBUG > 0
	printk("GPCM DEBUG: reset_ESCC() called for %s with dev=%p, dev->next=%p\n", dev->name, dev,
	       dev->next);
#endif

	cardbase = dev->base_addr & 0x3f0;

	save_flags(flags);
	cli();
	/* 
	 * Hardware reset
	 */
	wrtscc(cardbase, dev->base_addr + CTL, R9, FHWRES);
	/*
	 * Allow hardware reset completion
	 */
	udelay(20);
	/* 
	 * Disable interrupts with master interrupt ctrl reg
	 */
	wrtscc(cardbase, dev->base_addr + CTL, R9, 0);
	/* Disable IRQ & DMA request lines with card register */
	outb_p(DMAIRQOFF, cardbase + DMAIRQEN);
	/* 
	 * Set up counter chip timer 0 for 100 uS period square wave
	 * assuming a 4.9152 Mhz clock.
	 */
	outb_p(SC0 | LSB_MSB | MODE3, cardbase + TMRCMD);
	outb_p(DIV0 & 0xFF, cardbase + TMR0);
	outb_p(DIV0 >> 8, cardbase + TMR0);
	/* 
	 * Setup timer control word for timer 1 and 2
	 */
	outb_p(SC1 | LSB_MSB | MODE0, cardbase + TMRCMD);
	outb_p(0, cardbase + TMR1);
	outb_p(0, cardbase + TMR1);
	outb_p(SC2 | LSB_MSB | MODE0, cardbase + TMRCMD);
	outb_p(0, cardbase + TMR2);
	outb_p(0, cardbase + TMR2);

	restore_flags(flags);
}

/*
 * Init ESCC chip
 * This routine is called by gpcm_open()
 */
static void escc_init(struct device *dev)
{
	unsigned long flags;
	struct gpcm_local *lp = (struct gpcm_local *) dev->priv;

	int tc = 0;
	long br;
	register int cmd;

#if GPCM_DEBUG > 0
	printk("GPCM DEBUG: escc_init() called for %s with dev=%p, dev->next=%p\n", dev->name, dev,
	       dev->next);
#endif

	/* Initialize 85230 channel for SDLC operation */

	cmd = CTL + lp->base;
	save_flags(flags);
	cli();

	switch (cmd & CHANA) {
	case CHANA:
		wrtscc(lp->cardbase, cmd, R9, CHRA);	/* Reset channel A */
		wrtscc(lp->cardbase, cmd, R2, 0xff);	/* Initialize interrupt vector */
		break;
	default:
		wrtscc(lp->cardbase, cmd, R9, CHRB);	/* Reset channel B */
		break;
	}

	/* Deselect all Rx and Tx interrupts */
	wrtscc(lp->cardbase, cmd, R1, 0);

	/* Turn off external interrupts (like CTS/CD), select R7 */
	wrtscc(lp->cardbase, cmd, R15, FIFO_ENABLE);

	/* X1 clock, SDLC mode */
	wrtscc(lp->cardbase, cmd, R4, SDLC | X1CLK);

	/* Tx/Rx parameters */
	switch (lp->clockmode) {
	case INTERNAL:
		/* 
		 * Use internal clocking (38.4 Kbps max with 4.9 MHz oscillator)
		 * 
		 * Tx Clk from DPLL (=ESCC internal modulo 32 divisor).
		 * Rcv Clk from DPLL, TRxC pin outputs DPLL
		 */
		wrtscc(lp->cardbase, cmd, R10, CRCPS | NRZI);
		wrtscc(lp->cardbase, cmd, R11, TCDPLL | RCDPLL | TRxCDP | TRxCOI);
		break;
	case EXTERNAL:
		/* 
		 * Use external clocking 
		 *
		 * Tx Clk from Trxcl. Rcv Clk from Rtxcl, TRxC pin is input
		 */
		wrtscc(lp->cardbase, cmd, R10, CRCPS);
		wrtscc(lp->cardbase, cmd, R11, TCTRxCP);
		break;
		/* 
		 * WRITE-ME maybe another mode for G3RUH
		 * style modems?
		 */
	default:
		lp->clockmode = 0;
	}

	/* Null out SDLC start address */
	wrtscc(lp->cardbase, cmd, R6, 0);

	/* SDLC flag */
	wrtscc(lp->cardbase, cmd, R7, FLAG);

	/* 
	 * Set up the Transmitter but don't enable it
	 *  DTR, 8 bit TX chars only
	 */
	wrtscc(lp->cardbase, cmd, R5, Tx8 | DTR);

	/* Receiver initial setup */
	wrtscc(lp->cardbase, cmd, R3, Rx8);	/* 8 bits/char */

	/* Setting up BRG now - turn it off first */
	wrtscc(lp->cardbase, cmd, R14, BRSRC);	/* BRG off, keep Pclk source */

	/* set the 32x time constant for the BRG in Receive mode */

	switch (lp->clockmode) {
	case INTERNAL:
		/* internal clock */
		br = lp->speed;	/* get desired speed */
		tc = (lp->xtal / (64 * br)) - 2;	/* calc 32X BRG divisor */
#if GPCM_DEBUG > 1
		printk("GPCM: lp->speed=%d bps, tc=%d\n", lp->speed, tc);
#endif
		break;
	case EXTERNAL:
		tc = 14;
		break;
	}

	wrtscc(lp->cardbase, cmd, R12, tc & 0xFF);	/* lower byte */
	wrtscc(lp->cardbase, cmd, R13, (tc >> 8) & 0xFF);	/* upper byte */

	/* Following subroutine sets up and ENABLES the receiver */
	rts(lp, OFF);		/* TX OFF and RX ON */

	switch (lp->clockmode) {
	case INTERNAL:
		/* DPLL frm BRG, BRG src RTxC */
		wrtscc(lp->cardbase, cmd, R14, SSBR);
		wrtscc(lp->cardbase, cmd, R14, SEARCH);	/* SEARCH mode, keep BRG src */
		wrtscc(lp->cardbase, cmd, R14, BRENABL);	/* Enable the BRG */
		break;
	case EXTERNAL:
		/* DPLL frm RTxC, BRG src RTxC */
		wrtscc(lp->cardbase, cmd, R14, SSRTxC);
		wrtscc(lp->cardbase, cmd, R14, SEARCH);	/* SEARCH mode, keep BRG src */
		break;
	}

	if ((cmd & CHANA) == 0) {	/* if channel b */
		/*
		 * Interrupt on all characters, errors and external int 
		 */
		wrtscc(lp->cardbase, cmd, R1, (INT_ALL_Rx | EXT_INT_ENAB));
		/*
		 * Select R7'
		 */
		wrtscc(lp->cardbase, cmd, R15, ESCC | FIFO_ENABLE);
		/*
		 * Set port B RX & TX interrupt FIFO level
		 * TX int when TX FIFO empty
		 * RX int when RX FIFO half-full
		 */
		wrtscc(lp->cardbase, cmd, R7, RFIL | TFIL);
		wrtscc(lp->cardbase, cmd, R15, BRKIE | FIFO_ENABLE);	/* ABORT int */
	} else {
		/*
		 * Select R7'
		 */
		wrtscc(lp->cardbase, cmd, R15, ESCC);
		/*
		 * Set port A DMA request TX FIFO level
		 * 0    = when TX FIFO top byte empty
		 * TFIL = when TX FIFO completely empty
		 */
		wrtscc(lp->cardbase, cmd, R7, TXFIFO_LEVEL);
		wrtscc(lp->cardbase, cmd, R15, BRKIE | FIFO_ENABLE);
	}
	/* 
	 * Turn on the receiver and hunt for a flag 
	 * Use DCD-enabled RX to avoid tons of interrupts
	 */
	wrtscc(lp->cardbase, cmd, R3, RxENABLE | RxCRC_ENAB | Rx8 | AUTO_ENABLE);

	restore_flags(flags);
}


static void hardware_send_packet(struct gpcm_local *lp, struct sk_buff *skb)
{
	char kickflag;
	unsigned long flags;

	lp->stats.tx_packets++;

	save_flags(flags);
	cli();
	kickflag = (skb_peek(&lp->sndq) == NULL) && (lp->sndbuf == NULL);
	restore_flags(flags);

	skb_queue_tail(&lp->sndq, skb);

#if GPCM_DEBUG > 0
	printk("GPCM DEBUG: hardware_send_packet() called with kickflag=%d, state=%d\n", kickflag,
	       lp->tstate);
#endif
	if (kickflag) {
		/* simulate interrupt to xmit */
		switch (lp->base & CHANA) {
		case CHANA:
			if (lp->tstate == IDLE)
				a_txint(lp);	/* process interrupt */
			break;
		case CHANB:
			if (lp->tstate == IDLE)
				b_txint(lp);
			break;
		}
	}
}

static void setup_rx_dma(struct gpcm_local *lp)
{
	unsigned long flags;
	int cmd;
	unsigned long dma_abs;
	unsigned dmachan;

	save_flags(flags);
	cli();
	dma_abs = (unsigned long) (lp->rcvbuf->data);
	dmachan = lp->dmachan;
	cmd = lp->base + CTL;

	if (!valid_dma_page(dma_abs, DMA_BUFF_SIZE + sizeof(struct mbuf)))
		panic("GPCM: RX buffer violates DMA boundary!");

	/* Get ready for RX DMA */
	wrtscc(lp->cardbase, cmd, R1, WT_FN_RDYFN | WT_RDY_RT | INT_ERR_Rx | EXT_INT_ENAB);

	disable_dma(dmachan);
	clear_dma_ff(dmachan);

	/* Set DMA mode register to single transfers, incrementing address,
	 *  auto init, writes
	 */
	set_dma_mode(dmachan, DMA_MODE_READ | DMA_AUTOINIT);
	set_dma_addr(dmachan, dma_abs);
	set_dma_count(dmachan, lp->bufsiz);
	enable_dma(dmachan);

	/* If a packet is already coming in, this line is supposed to
	   avoid receiving a partial packet.
	 */
	wrtscc(lp->cardbase, cmd, R0, RES_Rx_CRC);

	/* Enable RX dma */
	wrtscc(lp->cardbase, cmd, R1,
	       WT_RDY_ENAB | WT_FN_RDYFN | WT_RDY_RT | INT_ERR_Rx | EXT_INT_ENAB);

	restore_flags(flags);
}

static void setup_tx_dma(struct gpcm_local *lp, int length)
{
	unsigned long dma_abs;
	unsigned long flags;
	unsigned long dmachan;

	save_flags(flags);
	cli();

	dmachan = lp->dmachan;
	dma_abs = (unsigned long) (lp->txdmabuf);

	if (!valid_dma_page(dma_abs, DMA_BUFF_SIZE + sizeof(struct mbuf)))
		panic("GPCM: TX buffer violates DMA boundary!");

	disable_dma(dmachan);
	/* Set DMA mode register to single transfers, incrementing address,
	 *  no auto init, reads
	 */
	set_dma_mode(dmachan, DMA_MODE_WRITE);
	clear_dma_ff(dmachan);
	set_dma_addr(dmachan, dma_abs);
	/* output byte count */
	set_dma_count(dmachan, length);

	restore_flags(flags);
}

static void tdelay(struct gpcm_local *lp, int time)
{
	int port;
	unsigned int t1;
	unsigned char sc;

	if (lp->base & CHANA) {	/* If A channel */
		sc = SC1;
		port = lp->cardbase + TMR1;
	} else {
		sc = SC2;
		port = lp->cardbase + TMR2;
		wrtscc(lp->cardbase, lp->base + CTL, R1, EXT_INT_ENAB);
	}

	t1 = time;

	/* Setup timer sc */
	outb_p(sc | LSB_MSB | MODE0, lp->cardbase + TMRCMD);
#ifdef IODELAY
	udelay(IODELAY);
#endif
	/* setup time delay in 100 us units */
	outb_p(t1 & 0xFF, port);
#ifdef IODELAY
	udelay(IODELAY);
#endif
	outb_p((t1 >> 8) & 0xFF, port);

	/* Enable correct int for timeout */
	wrtscc(lp->cardbase, lp->base + CTL, R15, CTSIE | FIFO_ENABLE);
	wrtscc(lp->cardbase, lp->base + CTL, R0, RES_EXT_INT);
}

static void free_p(struct sk_buff *skb)
{
	dev_kfree_skb(skb);
}

static inline void a_txint(struct gpcm_local *lp)
{
	int cmd;
	unsigned long flags;

	save_flags(flags);
	cli();

	cmd = CTL + lp->base;

#if GPCM_DEBUG > 0
	printk("GPCM DEBUG: a_txint() called, tstate=%d\n", lp->tstate);
#endif

	switch (lp->tstate) {
	case IDLE:
		/* Transmitter idle. Find a frame for transmission */
		if ((lp->sndbuf = skb_dequeue(&lp->sndq)) == NULL) {
			rts(lp, OFF);
			restore_flags(flags);
			return;
		}
		/* If a buffer to send, we drop thru here */
	case DEFER:
		/* we may have deferred prev xmit attempt */
		/* Check DCD - debounce it
		 * See Intel Microcommunications Handbook, p2-308
		 */
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		if ((rdscc(lp->cardbase, cmd, R0) & DCD) != 0) {
			lp->tstate = DEFER;
			tdelay(lp, 1000);
			/* defer until DCD transition or timeout */
			wrtscc(lp->cardbase, cmd, R15, CTSIE | DCDIE | FIFO_ENABLE);
			restore_flags(flags);
			return;
		}
		if (random() > lp->persist) {
			lp->tstate = DEFER;
			tdelay(lp, lp->slotime);
			restore_flags(flags);
			return;
		}
		rts(lp, ON);	/* Transmitter on */
		lp->tstate = ST_TXDELAY;
		tdelay(lp, lp->txdelay);
		restore_flags(flags);
		return;
	default:
		break;
	}			/* end switch(lp->state) */

	restore_flags(flags);
}				/*a_txint */

static inline void a_exint(struct gpcm_local *lp)
{
	unsigned long flags;
	int cmd;
	char st;
	int length;

	save_flags(flags);
	cli();			/* disable interrupts */

#if GPCM_DEBUG > 0
	printk("GPCM DEBUG: a_exint() called, tstate=%d, rstate=%d\n", lp->tstate, lp->rstate);
#endif
	st = rdscc(lp->cardbase, lp->base + CTL, R0);	/* Fetch status */

	/* reset external status latch */
	wrtscc(lp->cardbase, CTL + lp->base, R0, RES_EXT_INT);
	cmd = lp->base + CTL;

	if ((lp->rstate >= ACTIVE) && (st & BRK_ABRT)) {
		setup_rx_dma(lp);
		lp->rstate = ACTIVE;
	}
	switch (lp->tstate) {
	case ACTIVE:
		free_p(lp->sndbuf);
		lp->sndbuf = NULL;
		lp->tstate = FLAGOUT;
		tdelay(lp, lp->squeldelay);
		break;
	case FLAGOUT:
		if ((lp->sndbuf = skb_dequeue(&lp->sndq)) == NULL) {
			/* Nothing to send - return to receive mode */
			lp->tstate = IDLE;
			rts(lp, OFF);
			restore_flags(flags);
			return;
		}
		/* NOTE - fall through if more to send */
	case ST_TXDELAY:
		/* Disable DMA chan */
		disable_dma(lp->dmachan);

		/* Set up for TX dma */
		wrtscc(lp->cardbase, cmd, R1, WT_FN_RDYFN | EXT_INT_ENAB);

		/* Get all chars */

		/* Strip KISS control byte */
		length = lp->sndbuf->len - 1;

		memcpy(lp->txdmabuf, &lp->sndbuf->data[1], length);

		/* Setup DMA controller for tx */
		setup_tx_dma(lp, length);

		/* select transmit interrupts to enable */
		/* Allow DMA on chan */
		enable_dma(lp->dmachan);

		/* reset CRC, Txint pend */
		wrtscc(lp->cardbase, cmd, R0, RES_Tx_CRC | RES_Tx_P);

		/* allow Underrun int only */
		wrtscc(lp->cardbase, cmd, R15, TxUIE | FIFO_ENABLE);

		/* Enable TX DMA */
		wrtscc(lp->cardbase, cmd, R1, WT_RDY_ENAB | WT_FN_RDYFN | EXT_INT_ENAB);

		/* Send CRC on underrun */
		wrtscc(lp->cardbase, cmd, R0, RES_EOM_L);

		/* packet going out now */
		lp->tstate = ACTIVE;
		break;
	case DEFER:
		/* we have deferred prev xmit attempt
		 * See Intel Microcommunications Handbook, p2-308
		 */
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		if ((rdscc(lp->cardbase, cmd, R0) & DCD) != 0) {
			lp->tstate = DEFER;
			tdelay(lp, 1000);
			/* Defer until dcd transition or 100mS timeout */
			wrtscc(lp->cardbase, CTL + lp->base, R15, CTSIE | DCDIE | FIFO_ENABLE);
			restore_flags(flags);
			return;
		}
		if (random() > lp->persist) {
			lp->tstate = DEFER;
			tdelay(lp, lp->slotime);
			restore_flags(flags);
			return;
		}
		rts(lp, ON);	/* Transmitter on */
		lp->tstate = ST_TXDELAY;
		tdelay(lp, lp->txdelay);
		restore_flags(flags);
		return;
	}			/* switch(lp->tstate) */

	restore_flags(flags);
}				/* a_exint() */

/* Receive interrupt handler for the A channel
 */
static inline void a_rxint(struct device *dev, struct gpcm_local *lp)
{
	unsigned long flags;
	int cmd;
	int bytecount, bc1, bc2;
	char rse;
	struct sk_buff *skb;
	int sksize, pkt_len;
	struct mbuf *cur_buf;
	unsigned char *cfix;

	save_flags(flags);
	cli();			/* disable interrupts */
	cmd = lp->base + CTL;

#if GPCM_DEBUG > 0
	printk("GPCM DEBUG: a_rxint() called, rstate=%d\n", lp->rstate);
#endif

	/* bytecount from ESCC must be read before reading RR1 */
	bc1 = ((rdscc(lp->cardbase, cmd, R7) & 0x3f) << 8) + rdscc(lp->cardbase, cmd, R6);

	rse = rdscc(lp->cardbase, cmd, R1);	/* Get special condition bits from R1 */
	if (rse & Rx_OVR)
		lp->rstate = RXERROR;

	if (rse & END_FR) {
		/* If end of frame */
		clear_dma_ff(lp->dmachan);
		/* figure length of frame from 8237 */
		bc2 = lp->bufsiz - get_dma_residue(lp->dmachan);
#if GPCM_DEBUG > 1
		if (bc1 != bc2)
			printk("GPCM: ERROR: ESCC and DMA bytecounts does not agree (%d/%d)\n", bc1,
			       bc2);
#endif

#if FIFO_ENABLE == FIFOEN
		bytecount = bc1;	/* use ESCC value */
#else
		bytecount = bc2;	/* use DMA value */
#endif

#ifdef GPCM_DUMP_A
		{
			int n, i = 0;
			unsigned char *cp;

			for (cp = lp->rcvbuf->data, n = bytecount; n > 0; n--) {
				printk("%2.2x ", *cp++);
				if ((++i % 16) == 0)
					printk("\n");
			}
			printk("\n");
		}
#endif

		if ((rse & CRC_ERR) || (lp->rstate > ACTIVE) || (bytecount < 10)) {
			if ((bytecount >= 10) && (rse & CRC_ERR)) {
				lp->stats.rx_crc_errors++;
			}
			if (lp->rstate == RXERROR) {
				lp->stats.rx_errors++;
				lp->stats.rx_over_errors++;
			}
			/* Reset buffer pointers */
			lp->rstate = ACTIVE;
			setup_rx_dma(lp);
		} else {
			/* Here we have a valid frame */
#if GPCM_DEBUG > 0
			printk("Valid frame: bytecount from ESCC=%d, from DMA=%d\n", bc1, bc2);
#endif
			/* Toss 2 crc bytes , add one for KISS */
			pkt_len = lp->rcvbuf->cnt = bytecount - 2 + 1;

			/* Get buffer for next frame */
			cur_buf = lp->rcvbuf;
			switchbuffers(lp);
			setup_rx_dma(lp);

			/* Malloc up new buffer. */
			sksize = pkt_len;

			skb = dev_alloc_skb(sksize);
			if (skb == NULL) {
				printk("GPCM: %s: Memory squeeze, dropping packet.\n", dev->name);
				lp->stats.rx_dropped++;
				restore_flags(flags);
				return;
			}

			/* KISS kludge -  prefix with a 0 byte */
			cfix = skb_put(skb, pkt_len);
			*cfix++ = 0;

			/* 'skb->data' points to the start of sk_buff data area. */
			memcpy(cfix, (char *) cur_buf->data, pkt_len - 1);
			skb->dev = dev;
			skb->protocol = htons(ETH_P_AX25);
			skb->mac.raw = skb->data;
			netif_rx(skb);
			lp->stats.rx_packets++;
		}		/* end good frame */
	}			/* end EOF check */
	wrtscc(lp->cardbase, lp->base + CTL, R0, ERR_RES);	/* error reset */
	restore_flags(flags);
}

static inline void b_rxflush(struct gpcm_local *lp)
{
	int cmd = CTL + lp->base;
#if GPCM_DEBUG > 0
	register int f = 0;
	printk("b_rxflush called, state=%d\n", lp->rstate);
#endif
	lp->rcp = lp->rcvbuf->data;
	lp->rcp[0] = lp->rcvbuf->cnt = 0;

	while ((rdscc(lp->cardbase, cmd, R0)) & Rx_CH_AV) {
#if GPCM_DEBUG > 0
		++f;
#endif
		rdscc(lp->cardbase, cmd, R8);
	};
#if GPCM_DEBUG > 0
	printk("GPCM DEBUG: rxflush flushed %d bytes from fifo\n", f);
#endif
}

static inline void b_rxint(struct device *dev, struct gpcm_local *lp)
{
	unsigned long flags;
	int cmd;
	char rse;
	struct sk_buff *skb;
	int sksize;
	int pkt_len;
	unsigned char *cfix;

	save_flags(flags);
	cli();			/* disable interrupts */
	cmd = CTL + lp->base;

#if GPCM_DEBUG > 10
	printk("GPCM DEBUG: b_rxint() called, rstate=%d\n", lp->rstate);
#endif

	while ((rdscc(lp->cardbase, cmd, R0)) & Rx_CH_AV) {
		/* 
		 * Interrupt hits when there are at leat 4 chars in the FIFO
		 * read special condition bits before reading the data char
		 */
		rse = rdscc(lp->cardbase, cmd, R1);	/* get status byte from R1 */

		if (rse & Rx_OVR) {
			/* Rx overrun - toss buffer */
			/* reset buffer pointers */
			b_rxflush(lp);
			lp->rstate = RXERROR;	/* set error flag */
			lp->stats.rx_errors++;
			lp->stats.rx_over_errors++;
		} else if (lp->rcvbuf->cnt >= lp->bufsiz) {
			/* Too large -- toss buffer */
			/* reset buffer pointers */
			b_rxflush(lp);
			lp->rstate = TOOBIG;	/* when set, chars are not stored */
		}
		/* ok, we can store the received character now */
		if (lp->rstate == ACTIVE) {	/* If no errors... */
			*lp->rcp++ = rdscc(lp->cardbase, cmd, R8);
			/* char to rcv buff */
			lp->rcvbuf->cnt++;	/* bump count */
		} else {
			/* got to empty FIFO */
			while ((rdscc(lp->cardbase, cmd, R0)) & Rx_CH_AV) {
				rdscc(lp->cardbase, cmd, R8);
			}
			wrtscc(lp->cardbase, cmd, R0, ERR_RES);
			/* reset err latch */
			lp->rstate = ACTIVE;
		}
	}

	rse = rdscc(lp->cardbase, cmd, R1);	/* get status byte from R1 */

	if (rse & END_FR) {
		/* END OF FRAME -- Make sure Rx was active */
		if (lp->rcvbuf->cnt > 0) {
			if ((rse & CRC_ERR) || (lp->rstate > ACTIVE)
			    || (lp->rcvbuf->cnt < 10)) {
				if ((lp->rcvbuf->cnt >= 10) && (rse & CRC_ERR)) {
					lp->stats.rx_crc_errors++;
				}
				b_rxflush(lp);

			} else {
				/* Here we have a valid frame */
				pkt_len = lp->rcvbuf->cnt -= 2;	/* Toss 2 crc bytes */

				pkt_len += 1;	/* Make room for KISS control byte */
#ifdef GPCM_DUMP_B
				{
					int n, i = 0;
					unsigned char *cp;

					for (cp = lp->rcvbuf->data, n = lp->rcvbuf->cnt; n > 0; n--) {
						printk("%2.2x ", *cp++);
						if ((++i % 16) == 0)
							printk("\n");
					}
					printk("\n");
				}
#endif
				/* Malloc up new buffer. */
				sksize = pkt_len;
				skb = dev_alloc_skb(sksize);
				if (skb == NULL) {
					printk("GPCM: %s: Memory squeeze, dropping packet.\n",
					       dev->name);
					lp->stats.rx_dropped++;
					restore_flags(flags);
					return;
				}
				skb->dev = dev;

				/* KISS kludge -  prefix with a 0 byte */
				cfix = skb_put(skb, pkt_len);
				*cfix++ = 0;

				/* 'skb->data' points to the start of sk_buff data area. */
				memcpy(cfix, lp->rcvbuf->data, pkt_len - 1);
				skb->protocol = ntohs(ETH_P_AX25);
				skb->mac.raw = skb->data;
				netif_rx(skb);
				lp->stats.rx_packets++;
				/* packet queued - initialize buffer for next frame */
				b_rxflush(lp);


			}	/* end good frame queued */
		}		/* end check for active receive upon EOF */
		lp->rstate = ACTIVE;	/* and clear error status */
	}			/* end EOF check */
	restore_flags(flags);
}


static inline void b_txint(struct gpcm_local *lp)
{
	unsigned long flags;
	int cmd;
	unsigned char c;

	save_flags(flags);
	cli();
	cmd = CTL + lp->base;

#if GPCM_DEBUG > 10
	printk("GPCM DEBUG: b_txint() called, tstate=%d\n", lp->tstate);
#endif

	switch (lp->tstate) {
	case UNDERRUN:
		/* 
		 * stuffing more than 1 char per interrupt in TX FIFO
		 * cause another TX int to hit 2 to 4 character time after
		 * the frame end, when we have already switched to UNDERRUN
		 * state, so we must handle it here also.
		 */
		lp->tstate = CRCOUT;
		restore_flags(flags);
		return;
	case CRCOUT:
		lp->tstate = FLAGOUT;
		tdelay(lp, lp->squeldelay);
		restore_flags(flags);
		return;
	case IDLE:
		/* Transmitter idle. Find a frame for transmission */
		if ((lp->sndbuf = skb_dequeue(&lp->sndq)) == NULL) {
			/* Nothing to send - return to receive mode
			 * Tx OFF now - flag should have gone
			 */
			rts(lp, OFF);
			restore_flags(flags);
			return;
		}
		lp->txptr = lp->sndbuf->data;

		lp->txptr++;	/* Ignore KISS control byte */
		lp->txcnt = (int) lp->sndbuf->len - 1;

		/* If a buffer to send, we drop thru here */
	case DEFER:		/* we may have deferred prev xmit attempt */
		/* Check DCD - debounce it */
		/* See Intel Microcommunications Handbook, p2-308 */
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		if ((rdscc(lp->cardbase, cmd, R0) & DCD) != 0) {
			lp->tstate = DEFER;
			tdelay(lp, 1000);
			/* defer until DCD transition or timeout */
			wrtscc(lp->cardbase, cmd, R15, CTSIE | DCDIE | FIFO_ENABLE);
			restore_flags(flags);
			return;
		}
		if (random() > lp->persist) {
			lp->tstate = DEFER;
			tdelay(lp, lp->slotime);
			restore_flags(flags);
			return;
		}
		rts(lp, ON);	/* Transmitter on */
		lp->tstate = ST_TXDELAY;
		tdelay(lp, lp->txdelay);
		restore_flags(flags);
		return;
	case ACTIVE:
		/* 
		 * Here we are actively sending a frame
		 * Continue writing characters until TX FIFO full
		 * Next interrupt will hit when TX FIFO is empty again
		 */
#if GPCM_DEBUG > 2
		printk("+");
#endif
		while (rdscc(lp->cardbase, cmd, R0) & Tx_BUF_EMP) {	/* If TX FIFO not full */
			if (lp->txcnt > 0) {	/* If there are more characters to send */
				/* 
				 * Read and send another char
				 */
				c = *lp->txptr++;
				lp->txcnt--;
				wrtscc(lp->cardbase, cmd, R8, c);
				/* Stuffing a char satisfies Interrupt condition */
			} else {
				/* 
				 * No more characters to send, so exit anyway,
				 * even if TX FIFO not full
				 */
				break;
			}
		}
		if (lp->txcnt == 0) {
#if GPCM_DEBUG > 2
			printk("#");
#endif
			/* No more to send */
			free_p(lp->sndbuf);
			lp->sndbuf = NULL;
#if 0
			/* 
			 * Is that really needed ?
			 * An underrun does not automatically triggers
			 * an EXTERNAL INTERRUPT, so why check it here ?
			 */
			if ((rdscc(lp->cardbase, cmd, R0) & TxEOM)) {
#if GPCM_DEBUG > 2
				printk("!");
#endif
				/* Did we underrun? */
				/* unexpected underrun */
				lp->stats.tx_errors++;
				lp->stats.tx_fifo_errors++;
				wrtscc(lp->cardbase, cmd, R0, SEND_ABORT);
				lp->tstate = FLAGOUT;
				tdelay(lp, lp->squeldelay);
				restore_flags(flags);
				return;
			}
#endif
			lp->tstate = UNDERRUN;	/* Now we expect to underrun */
			/* Send flags on underrun */
			switch (lp->clockmode) {
			case INTERNAL:
				/* If internally clocked */
				wrtscc(lp->cardbase, cmd, R10, CRCPS | NRZI);
				break;
			case EXTERNAL:
				wrtscc(lp->cardbase, cmd, R10, CRCPS);
				break;
			}
			wrtscc(lp->cardbase, cmd, R0, RES_Tx_P);	/* reset Tx Int Pend */
#if GPCM_DEBUG > 2
			printk(".");
#endif
		}
		restore_flags(flags);
		return;		/* back to wait for interrupt */
#if GPCM_DEBUG > 0
	default:
		printk("b_txint() called with unknown state %d\n", lp->tstate);
		break;
#endif
	}			/* end switch */
	restore_flags(flags);
}

/* 
 * GPCM SIO External/Status interrupts (for the B channel)
 * This can be caused by a receiver abort, or a Tx UNDERRUN/EOM.
 * Receiver automatically goes to Hunt on an abort.
 *
 * If the Tx Underrun interrupt hits, change state and
 * issue a reset command for it, and return.
 */
static inline void b_exint(struct gpcm_local *lp)
{
	unsigned long flags;
	char st;
	int cmd, n;
	char c;

	cmd = CTL + lp->base;
	save_flags(flags);
	cli();			/* disable interrupts */
	st = rdscc(lp->cardbase, cmd, R0);	/* Fetch status */
	/* reset external status latch */
	wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);

#if GPCM_DEBUG > 10
	printk("GPCM DEBUG: b_exint() called, tstate=%d, rstate=%d\n", lp->tstate, lp->rstate);
#endif

	switch (lp->tstate) {
	case ACTIVE:		/* Unexpected underrun */
#if FPCM_DEBUG > 0
		printk("Port B TX underrun when %d chars remaining\n", lp->txcnt);
#endif
		free_p(lp->sndbuf);
		lp->sndbuf = NULL;
		wrtscc(lp->cardbase, cmd, R0, SEND_ABORT);
		lp->tstate = FLAGOUT;
		lp->stats.tx_errors++;
		lp->stats.tx_fifo_errors++;
		tdelay(lp, lp->squeldelay);
		restore_flags(flags);
		return;
	case UNDERRUN:
		lp->tstate = CRCOUT;
		restore_flags(flags);
		return;
	case FLAGOUT:
		/* Find a frame for transmission */
		if ((lp->sndbuf = skb_dequeue(&lp->sndq)) == NULL) {
			/* Nothing to send - return to receive mode
			 * Tx OFF now - flag should have gone
			 */
			rts(lp, OFF);
			lp->tstate = IDLE;
			restore_flags(flags);
			return;
		}
		lp->txptr = lp->sndbuf->data;

		lp->txptr++;	/* Ignore KISS control byte */
		lp->txcnt = (int) lp->sndbuf->len - 1;

		wrtscc(lp->cardbase, cmd, R0, RES_Tx_CRC);	/* reset for next frame */

		/* Send abort on underrun */
		switch (lp->clockmode) {
		case INTERNAL:
			/* If internally clocked */
			wrtscc(lp->cardbase, cmd, R10, CRCPS | NRZI | ABUNDER);
			break;
		case EXTERNAL:
			wrtscc(lp->cardbase, cmd, R10, CRCPS | ABUNDER);
			break;
		}
		/*
		 * Fill out transmit FIFO (works only with 85230 4-byte FIFO)
		 */
		for (n = 0; n < 4; n++) {
			lp->txcnt--;
			c = *lp->txptr++;
			wrtscc(lp->cardbase, cmd, R8, c);
		}

		wrtscc(lp->cardbase, cmd, R0, RES_EOM_L);	/* Reset end of message latch */

		/* select transmit interrupts to enable */
		wrtscc(lp->cardbase, cmd, R15, TxUIE | FIFO_ENABLE);	/* allow Underrun int only */
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		wrtscc(lp->cardbase, cmd, R1, TxINT_ENAB | EXT_INT_ENAB);	/* Tx/Ext ints */
		lp->tstate = ACTIVE;	/* char going out now */
		restore_flags(flags);
		return;
	case DEFER:
		/* Check DCD - debounce it
		 * See Intel Microcommunications Handbook, p2-308
		 */
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		if ((rdscc(lp->cardbase, cmd, R0) & DCD) != 0) {
			lp->tstate = DEFER;
			tdelay(lp, 1000);
			/* defer until DCD transition or timeout */
			wrtscc(lp->cardbase, cmd, R15, CTSIE | DCDIE | FIFO_ENABLE);
			restore_flags(flags);
			return;
		}
		if (random() > lp->persist) {
			lp->tstate = DEFER;
			tdelay(lp, lp->slotime);
			restore_flags(flags);
			return;
		}
		rts(lp, ON);	/* Transmitter on */
		lp->tstate = ST_TXDELAY;
		tdelay(lp, lp->txdelay);
		restore_flags(flags);
		return;
	case ST_TXDELAY:
		wrtscc(lp->cardbase, cmd, R0, RES_Tx_CRC);	/* reset for next frame */

		/* Send abort on underrun */
		switch (lp->clockmode) {
		case INTERNAL:
			/* If internally clocked */
			wrtscc(lp->cardbase, cmd, R10, CRCPS | NRZI | ABUNDER);
			break;
		case EXTERNAL:
			wrtscc(lp->cardbase, cmd, R10, CRCPS | ABUNDER);
			break;
		}
		/*
		 * Fill out transmit FIFO (works only with 85230 4-byte FIFO)
		 */
		for (n = 0; n < 4; n++) {
			lp->txcnt--;
			c = *lp->txptr++;
			wrtscc(lp->cardbase, cmd, R8, c);
		}

		wrtscc(lp->cardbase, cmd, R0, RES_EOM_L);	/* Reset end of message latch */

		/* select transmit interrupts to enable */
		wrtscc(lp->cardbase, cmd, R15, TxUIE | FIFO_ENABLE);	/* allow Underrun int only */
		wrtscc(lp->cardbase, cmd, R0, RES_EXT_INT);
		/* Tx/Extern ints on */
		wrtscc(lp->cardbase, cmd, R1, TxINT_ENAB | EXT_INT_ENAB);
		lp->tstate = ACTIVE;	/* char going out now */
#if GPCM_DEBUG > 2
		printk("*");
#endif
		restore_flags(flags);
		return;
#if GPCM_DEBUG > 0
	default:
		printk("b_exint() called with unknown state %d\n", lp->tstate);
		break;
#endif
	}

	/* Receive Mode only
	 * This triggers when hunt mode is entered, & since an ABORT
	 * automatically enters hunt mode, we use that to clean up
	 * any waiting garbage
	 */
	if ((lp->rstate == ACTIVE) && (st & BRK_ABRT)) {
		while ((rdscc(lp->cardbase, cmd, R0)) & Rx_CH_AV) {
			(void) rdscc(lp->cardbase, cmd, R8);
		}
		b_rxflush(lp);

	}
	restore_flags(flags);
}				/*b_exint */

static void rts(struct gpcm_local *lp, int x)
{
	int cmd;

	/* assumes interrupts are off */
	cmd = CTL + lp->base;

	/* Turn on transmitter to send flags */
	if (x == ON) {
		/* Turn Tx ON and Receive OFF */
		/* Exints off first to avoid abort int */
		wrtscc(lp->cardbase, cmd, R15, FIFO_ENABLE);
		wrtscc(lp->cardbase, cmd, R3, Rx8);	/* Rx off, no auto enable on CTS */
		lp->rstate = IDLE;
		if (cmd & CHANA) {	/* if channel A */
			/* Set up for TX dma */
			wrtscc(lp->cardbase, cmd, R1, WT_FN_RDYFN | EXT_INT_ENAB);
		} else {
			wrtscc(lp->cardbase, cmd, R1, 0);	/* No interrupts */
		}
		wrtscc(lp->cardbase, cmd, R5, TxCRC_ENAB | RTS | TxENAB | Tx8 | DTR);
		/* Transmitter now on */
	} else {
		/* Tx OFF and Rx ON */
		lp->tstate = IDLE;
		wrtscc(lp->cardbase, cmd, R5, Tx8 | DTR);	/*  TX off */
		/* Now, turn on the receiver, set auto enable on DCD and hunt for a flag */
		wrtscc(lp->cardbase, cmd, R3, RxENABLE | RxCRC_ENAB | Rx8 | AUTO_ENABLE);
		lp->rstate = ACTIVE;	/* Normal state */

		if (cmd & CHANA) {	/* if channel A */
			setup_rx_dma(lp);
		} else {
			/* reset buffer pointers */
			b_rxflush(lp);

			wrtscc(lp->cardbase, cmd, R1, (INT_ALL_Rx | EXT_INT_ENAB));
		}
		wrtscc(lp->cardbase, cmd, R15, BRKIE | FIFO_ENABLE);	/* allow ABORT int */
	}
}

static int valid_dma_page(unsigned long addr, unsigned long dev_buffsize)
{
	if (((addr & 0xffff) + dev_buffsize) <= 0x10000)
		return 1;
	else
		return 0;
}

static int gpcm_set_mac_address(struct device *dev, void *addr)
{
	struct sockaddr *sa = (struct sockaddr *) addr;
	memcpy(dev->dev_addr, sa->sa_data, dev->addr_len);	/* addr is an AX.25 shifted ASCII */
	return 0;		/* mac address */
}

/* Allocate a buffer which does not cross a DMA page boundary */
static char *get_dma_buffer(unsigned long *mem_ptr)
{
	char *ret;

	ret = (char *) *mem_ptr;

	if (!valid_dma_page(*mem_ptr, DMA_BUFF_SIZE + sizeof(struct mbuf))) {
		*mem_ptr += (DMA_BUFF_SIZE + sizeof(struct mbuf));
		ret = (char *) *mem_ptr;
	}
	*mem_ptr += (DMA_BUFF_SIZE + sizeof(struct mbuf));
	return (ret);
}

static int gpcm_send_packet(struct sk_buff *skb, struct device *dev)
{
	struct gpcm_local *lp = (struct gpcm_local *) dev->priv;

	hardware_send_packet(lp, skb);
	dev->trans_start = jiffies;

	return 0;
}

/* 
 * The typical workload of the driver:
 * Handle the network interface interrupts.
 */
static void gpcm_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int st;
	struct gpcm_local *lp;
	unsigned long flags;
	struct device *dev_a, *dev_b;

	dev_a = &dev_gpcm0a;
	dev_b = &dev_gpcm0b;
	/*
	 * Read interrupt status register (only valid from channel A).
	 * Process all pending interrupts in while loop.
	 *
	 * The if order MUST MATCH with the interrupts priority list;
	 * otherwise, the RES_H_IUS command issued at the end may
	 * reset THE WRONG INTERRUPT and the while will loop forever !!
	 *
	 * From Zilog 85230 data sheet the 85230 interrupt priority is
	 * (from highest to lowest):
	 * Channel A RX, Channel A TX, Channel A EXT,
	 * Channel B RX, Channel B TX, Channel B EXT.
	 */
	lp = (struct gpcm_local *) dev_a->priv;	/* Assume channel A */
	while ((st = rdscc(lp->cardbase, dev_a->base_addr | CHANA | CTL, R3)) != 0) {
		if (st & CHARxIP) {
			/* Channel A Receive Interrupt Pending */
			lp = (struct gpcm_local *) dev_a->priv;
			a_rxint(dev_a, lp);
		} else if (st & CHATxIP) {
			/* Channel A Transmit Interrupt Pending */
			lp = (struct gpcm_local *) dev_a->priv;
			a_txint(lp);
		} else if (st & CHAEXT) {
			/* Channel A External Status Interrupt Pending */
			lp = (struct gpcm_local *) dev_a->priv;
			a_exint(lp);
		} else if (st & CHBRxIP) {
			/* Channel A Receive Interrupt Pending */
			lp = (struct gpcm_local *) dev_b->priv;
			b_rxint(dev_b, lp);
		} else if (st & CHBTxIP) {
			/* Channel B Rcv Interrupt Pending */
			lp = (struct gpcm_local *) dev_b->priv;
			b_txint(lp);
		} else if (st & CHBEXT) {
			/* Channel B External Status Int */
			lp = (struct gpcm_local *) dev_b->priv;
			b_exint(lp);
		}
		/* Reset highest interrupt under service */

		save_flags(flags);
		cli();
		wrtscc(lp->cardbase, lp->base + CTL, R0, RES_H_IUS);
		restore_flags(flags);
	}			/* End of while loop on int processing */
	return;
}

/* 
 * Open/initialize the board.  This is called (in the current kernel)
 * sometime after booting when the 'ifconfig' program is run.
 *
 * This routine should set everything up anew at each open, even
 * registers that "should" only need to be set once at boot, so that
 * there is non-reboot way to recover if something goes wrong.
 */
static int gpcm_open(struct device *dev)
{
	unsigned long flags;

	struct gpcm_local *lp = (struct gpcm_local *) dev->priv;

#if GPCM_DEBUG > 0
	printk("gpcm_open() called for %s with dev=%p, dev->next=%p\n", dev->name, dev, dev->next);
	printk("             add=0x%lx, irq=%u, dma=%u\n", dev->base_addr, dev->irq, dev->dma);
#endif

	if (dev->base_addr & CHANA) {	/* if channel A */
		if (dev->start == 0) {	/* not started yet, grab IRQ & DMA */
			int irqval = request_irq(dev->irq, &gpcm_interrupt, 0,
						 "gpcm", dev);
			if (irqval) {
				printk("GPCM: unable to get IRQ %d (irqval=%d).\n", dev->irq,
				       irqval);
				return -EAGAIN;
			}

			if (request_dma(dev->dma, "gpcm")) {
				free_irq(dev->irq, dev);
				return -EAGAIN;
			}
		}
		/* Reset the hardware here. */
		reset_ESCC(dev);
	}
	lp->tstate = IDLE;

	escc_init(dev);

	/* master interrupt enable */
	save_flags(flags);
	cli();
	wrtscc(lp->cardbase, CTL + lp->base, R9, MIE | NV);
	restore_flags(flags);

	lp->open_time = jiffies;

	dev->tbusy = 0;
	dev->interrupt = 0;
	dev->start = 1;

	MOD_INC_USE_COUNT;

	return 0;

}

/* 
 * The inverse routine to gpcm_open().
 */
static int gpcm_close(struct device *dev)
{
	unsigned long flags;
	struct gpcm_local *lp;
	struct sk_buff *ptr;

#if GPCM_DEBUG > 0
	printk("gpcm_close() called for %s with dev=%p, dev->next=%p\n", dev->name, dev, dev->next);
#endif
	save_flags(flags);
	cli();

	lp = (struct gpcm_local *) dev->priv;
	ptr = NULL;

	reset_ESCC(dev);	/* reset both ESCC channels */
	disable_dma(lp->dmachan);

	lp->open_time = 0;

	dev->tbusy = 1;
	dev->start = 0;

	/* Free any buffers left in the hardware transmit queue */
	while ((ptr = skb_dequeue(&lp->sndq)) != NULL)
		kfree_skb(ptr);

	if (dev->base_addr & CHANA) {
		/* Release IRQ & DMA channel */
		outb_p(DMAIRQOFF, lp->cardbase + DMAIRQEN);
		free_irq(dev->irq, dev);	/* IRQs and IO Ports are shared */
		free_dma(dev->dma);	/* Release DMA channel */
	}
	restore_flags(flags);

	MOD_DEC_USE_COUNT;

	return 0;
}

static int gpcm_ioctl(struct device *dev, struct ifreq *ifr, int cmd)
{
	unsigned long flags;
	struct gpcm_req rq;
	struct gpcm_local *lp = (struct gpcm_local *) dev->priv;
	struct device *dev_a = &dev_gpcm0a;
	struct device *dev_b = &dev_gpcm0b;

	int ret = verify_area(VERIFY_WRITE, ifr->ifr_data,
			      sizeof(struct gpcm_req));
	if (ret)
		return ret;

	if (cmd != SIOCDEVPRIVATE)
		return -EINVAL;

	copy_from_user(&rq, ifr->ifr_data, sizeof(struct gpcm_req));

	switch (rq.cmd) {
	case SIOCSGPCMPARAM:
		if (!suser())
			return -EPERM;
		if (rq.speed > 38400)
			return -EINVAL;
		if (rq.clockmode >= NUM_CLOCK_MODES)
			return -EINVAL;
		save_flags(flags);
		cli();
		lp->txdelay = rq.txdelay;
		lp->persist = rq.persist;
		lp->slotime = rq.slotime;
		lp->squeldelay = rq.squeldelay;
		lp->clockmode = rq.clockmode;
		lp->speed = rq.speed;
		if (dev_a != NULL) {
			gpcm_open(dev_a);	/* MOD_INC_USE_COUNT executed here */
			gpcm_open(dev_b);	/* ditto */

			MOD_DEC_USE_COUNT;
			MOD_DEC_USE_COUNT;

		}
		restore_flags(flags);
		ret = 0;
		break;
	case SIOCSGPCMDMA:
	case SIOCSGPCMIRQ:
		/* 
		 * DMA & IRQ are set by insmod command-line parameters,
		 * No need to set them through IOCTL's.
		 */
		ret = -EINVAL;
		break;
	case SIOCGGPCMPARAM:
	case SIOCGGPCMDMA:
	case SIOCGGPCMIRQ:
		rq.speed = lp->speed;
		rq.txdelay = lp->txdelay;
		rq.persist = lp->persist;
		rq.slotime = lp->slotime;
		rq.squeldelay = lp->squeldelay;
		rq.clockmode = lp->clockmode;
		rq.dmachan = lp->dmachan;
		rq.irq = dev->irq;
		copy_to_user(ifr->ifr_data, &rq, sizeof(struct gpcm_req));
		ret = 0;
		break;
	default:
		ret = -EINVAL;
	}
	return ret;
}

/*
 * Get the current statistics.  This may be called with the card open or closed.
 */
static struct netstats *gpcm_get_stats(struct device *dev)
{
	struct gpcm_local *lp = (struct gpcm_local *) dev->priv;

	return &lp->stats;
}

MODULE_AUTHOR("I1EPJ (i1epj@libero.it) & IW1CFL ()");
MODULE_DESCRIPTION("GPCM ESCC card DMA driver v0.8");

static int io = 0;
static int irq = 0;
static int dma = 0;

MODULE_PARM(io, "1i");
MODULE_PARM(irq, "1i");
MODULE_PARM(dma, "1i");

int init_module(void)
{
	dev_gpcm0a.base_addr = io + 2;
	dev_gpcm0a.irq = irq;
	dev_gpcm0a.dma = dma;

	if (io == 0 || irq == 0 || dma == 0) {
		printk("GPCM: ERROR: must supply io, irq and dma parameters.\n");
		return -EIO;
	}
	if (register_netdev(&dev_gpcm0a) != 0) {
		printk(KERN_INFO "gpcm0a: not registered\n");
		return -EIO;
	};

	dev_gpcm0b.base_addr = io;
	dev_gpcm0b.irq = dev_gpcm0a.irq;
	dev_gpcm0b.dma = 0;

	if (register_netdev(&dev_gpcm0b) != 0) {
		printk(KERN_INFO "gpcm0b: not registered\n");
		return -EIO;
	};
	return 0;
}

void cleanup_module(void)
{
	/* No need to check MOD_IN_USE, as sys_delete_module() checks. */

	release_region(dev_gpcm0a.base_addr & 0x3f0, GPCM_TOTAL_SIZE);

	if (dev_gpcm0a.priv != NULL)
		kfree(dev_gpcm0a.priv);
	dev_gpcm0a.priv = NULL;
	unregister_netdev(&dev_gpcm0a);

	dev_gpcm0b.irq = 0;

	if (dev_gpcm0b.priv != NULL)
		kfree(dev_gpcm0b.priv);
	dev_gpcm0b.priv = NULL;
	unregister_netdev(&dev_gpcm0b);

}

#endif				/* ifndef MODULE */
